;
; ***** BEGIN LICENSE BLOCK *****
; Version: MPL 1.1/GPL 2.0/LGPL 2.1
; 
; The contents of this file are subject to the Mozilla Public License Version 
; 1.1 (the "License"); you may not use this file except in compliance with 
; the License. You may obtain a copy of the License at 
; http://www.mozilla.org/MPL/
; 
; Software distributed under the License is distributed on an "AS IS" basis,
; WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
; for the specific language governing rights and limitations under the
; License.
; 
; The Original Code is Mozilla Communicator client code.
; 
; The Initial Developer of the Original Code is
; Netscape Communications Corporation.
; Portions created by the Initial Developer are Copyright (C) 1996-1999
; the Initial Developer. All Rights Reserved.
; 
; Contributor(s):
; 
; Alternatively, the contents of this file may be used under the terms of
; either of the GNU General Public License Version 2 or later (the "GPL"),
; or the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
; in which case the provisions of the GPL or the LGPL are applicable instead
; of those above. If you wish to allow use of your version of this file only
; under the terms of either the GPL or the LGPL, and not to allow others to
; use your version of this file under the terms of the MPL, indicate your
; decision by deleting the provisions above and replace them with the notice
; and other provisions required by the GPL or the LGPL. If you do not delete
; the provisions above, a recipient may use your version of this file under
; the terms of any one of the MPL, the GPL or the LGPL.
; 
; ***** END LICENSE BLOCK ***** 

LIBRARY         NSLDAPPR32V60
DESCRIPTION     'LDAP to NSPR glue library for 32-bit Windows'
VERSION		6.0
HEAPSIZE        4096

EXPORTS
; we need to manually assign ordinal numbers so we can add new routines
; and not disturb the ordinals and thus not require callers to relink.
; 
; exports list (generated by genexports.pl)
; 

	prldap_init		@2010
	prldap_install_routines		@2011
	prldap_set_session_info		@2012
	prldap_get_session_info		@2013
	prldap_set_socket_info		@2014
	prldap_get_socket_info		@2015
	prldap_set_session_option		@2016
	prldap_get_session_option		@2017
	prldap_is_installed		@2018
	prldap_import_connection		@2019
	prldap_set_default_socket_info		@2020
	prldap_get_default_socket_info		@2021
; 
; end of generated exports list.
